@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2010, 2011, 2012, 2013, 2014 Nicira, Inc.
+ * Copyright (c) 2010, 2011, 2012, 2013, 2014, 2017 Nicira, Inc.
  * Copyright (c) 2016 Red Hat, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -45,6 +45,10 @@
 #include "unaligned.h"
 #include "unixctl.h"
 #include "openvswitch/vlog.h"
+#include "netdev-tc-offloads.h"
+#ifdef __linux__
+#include "netdev-linux.h"
+#endif
 
 VLOG_DEFINE_THIS_MODULE(netdev_vport);
 
@@ -94,14 +98,6 @@ netdev_vport_is_patch(const struct netdev *netdev)
     return class->get_config == get_patch_config;
 }
 
-bool
-netdev_vport_is_layer3(const struct netdev *dev)
-{
-    const char *type = netdev_get_type(dev);
-
-    return (!strcmp("lisp", type));
-}
-
 static bool
 netdev_vport_needs_dst_port(const struct netdev *dev)
 {
@@ -260,10 +256,12 @@ tunnel_check_status_change__(struct netdev_vport *netdev)
     bool status = false;
     struct in6_addr *route;
     struct in6_addr gw;
+    uint32_t mark;
 
     iface[0] = '\0';
     route = &netdev->tnl_cfg.ipv6_dst;
-    if (ovs_router_lookup(route, iface, NULL, &gw)) {
+    mark = netdev->tnl_cfg.egress_pkt_mark;
+    if (ovs_router_lookup(mark, route, iface, NULL, &gw)) {
         struct netdev *egress_netdev;
 
         if (!netdev_open(iface, NULL, &egress_netdev)) {
@@ -274,7 +272,7 @@ tunnel_check_status_change__(struct netdev_vport *netdev)
 
     if (strcmp(netdev->egress_iface, iface)
         || netdev->carrier_status != status) {
-        ovs_strlcpy(netdev->egress_iface, iface, IFNAMSIZ);
+        ovs_strlcpy_arrays(netdev->egress_iface, iface);
         netdev->carrier_status = status;
 
         return true;
@@ -397,6 +395,31 @@ parse_tunnel_ip(const char *value, bool accept_mcast, bool *flow,
     return 0;
 }
 
+enum tunnel_layers {
+    TNL_L2 = 1 << 0,       /* 1 if a tunnel type can carry Ethernet traffic. */
+    TNL_L3 = 1 << 1        /* 1 if a tunnel type can carry L3 traffic. */
+};
+static enum tunnel_layers
+tunnel_supported_layers(const char *type,
+                        const struct netdev_tunnel_config *tnl_cfg)
+{
+    if (!strcmp(type, "lisp")) {
+        return TNL_L3;
+    } else if (!strcmp(type, "gre")) {
+        return TNL_L2 | TNL_L3;
+    } else if (!strcmp(type, "vxlan")
+               && tnl_cfg->exts & (1 << OVS_VXLAN_EXT_GPE)) {
+        return TNL_L2 | TNL_L3;
+    } else {
+        return TNL_L2;
+    }
+}
+static enum netdev_pt_mode
+default_pt_mode(enum tunnel_layers layers)
+{
+    return layers == TNL_L3 ? NETDEV_PT_LEGACY_L3 : NETDEV_PT_LEGACY_L2;
+}
+
 static int
 set_tunnel_config(struct netdev *dev_, const struct smap *args, char **errp)
 {
@@ -488,9 +511,10 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args, char **errp)
             }
         } else if (!strcmp(node->key, "key") ||
                    !strcmp(node->key, "in_key") ||
-                   !strcmp(node->key, "out_key")) {
+                   !strcmp(node->key, "out_key") ||
+                   !strcmp(node->key, "packet_type")) {
             /* Handled separately below. */
-        } else if (!strcmp(node->key, "exts")) {
+        } else if (!strcmp(node->key, "exts") && !strcmp(type, "vxlan")) {
             char *str = xstrdup(node->value);
             char *ext, *save_ptr = NULL;
 
@@ -500,6 +524,8 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args, char **errp)
             while (ext) {
                 if (!strcmp(type, "vxlan") && !strcmp(ext, "gbp")) {
                     tnl_cfg.exts |= (1 << OVS_VXLAN_EXT_GBP);
+                } else if (!strcmp(type, "vxlan") && !strcmp(ext, "gpe")) {
+                    tnl_cfg.exts |= (1 << OVS_VXLAN_EXT_GPE);
                 } else {
                     ds_put_format(&errors, "%s: unknown extension '%s'\n",
                                   name, ext);
@@ -509,12 +535,49 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args, char **errp)
             }
 
             free(str);
+        } else if (!strcmp(node->key, "egress_pkt_mark")) {
+            tnl_cfg.egress_pkt_mark = strtoul(node->value, NULL, 10);
+            tnl_cfg.set_egress_pkt_mark = true;
         } else {
-            ds_put_format(&errors, "%s: unknown %s argument '%s'\n",
-                          name, type, node->key);
+            ds_put_format(&errors, "%s: unknown %s argument '%s'\n", name,
+                          type, node->key);
         }
     }
 
+    enum tunnel_layers layers = tunnel_supported_layers(type, &tnl_cfg);
+    const char *full_type = (strcmp(type, "vxlan") ? type
+                             : (tnl_cfg.exts & (1 << OVS_VXLAN_EXT_GPE)
+                                ? "VXLAN-GPE" : "VXLAN (without GPE"));
+    const char *packet_type = smap_get(args, "packet_type");
+    if (!packet_type) {
+        tnl_cfg.pt_mode = default_pt_mode(layers);
+    } else if (!strcmp(packet_type, "legacy_l2")) {
+        tnl_cfg.pt_mode = NETDEV_PT_LEGACY_L2;
+        if (!(layers & TNL_L2)) {
+            ds_put_format(&errors, "%s: legacy_l2 configured on %s tunnel "
+                          "that cannot carry L2 traffic\n",
+                          name, full_type);
+            err = EINVAL;
+            goto out;
+        }
+    } else if (!strcmp(packet_type, "legacy_l3")) {
+        tnl_cfg.pt_mode = NETDEV_PT_LEGACY_L3;
+        if (!(layers & TNL_L3)) {
+            ds_put_format(&errors, "%s: legacy_l3 configured on %s tunnel "
+                          "that cannot carry L3 traffic\n",
+                          name, full_type);
+            err = EINVAL;
+            goto out;
+        }
+    } else if (!strcmp(packet_type, "ptap")) {
+        tnl_cfg.pt_mode = NETDEV_PT_AWARE;
+    } else {
+        ds_put_format(&errors, "%s: unknown packet_type '%s'\n",
+                      name, packet_type);
+        err = EINVAL;
+        goto out;
+    }
+
     if (!ipv6_addr_is_set(&tnl_cfg.ipv6_dst) && !tnl_cfg.ip_dst_flow) {
         ds_put_format(&errors,
                       "%s: %s type requires valid 'remote_ip' argument\n",
@@ -578,6 +641,7 @@ static int
 get_tunnel_config(const struct netdev *dev, struct smap *args)
 {
     struct netdev_vport *netdev = netdev_vport_cast(dev);
+    const char *type = netdev_get_type(dev);
     struct netdev_tunnel_config tnl_cfg;
 
     ovs_mutex_lock(&netdev->mutex);
@@ -631,7 +695,6 @@ get_tunnel_config(const struct netdev *dev, struct smap *args)
 
     if (tnl_cfg.dst_port) {
         uint16_t dst_port = ntohs(tnl_cfg.dst_port);
-        const char *type = netdev_get_type(dev);
 
         if ((!strcmp("geneve", type) && dst_port != GENEVE_DST_PORT) ||
             (!strcmp("vxlan", type) && dst_port != VXLAN_DST_PORT) ||
@@ -645,10 +708,22 @@ get_tunnel_config(const struct netdev *dev, struct smap *args)
         smap_add(args, "csum", "true");
     }
 
+    enum tunnel_layers layers = tunnel_supported_layers(type, &tnl_cfg);
+    if (tnl_cfg.pt_mode != default_pt_mode(layers)) {
+        smap_add(args, "packet_type",
+                 tnl_cfg.pt_mode == NETDEV_PT_LEGACY_L2 ? "legacy_l2"
+                 : tnl_cfg.pt_mode == NETDEV_PT_LEGACY_L3 ? "legacy_l3"
+                 : "ptap");
+    }
+
     if (!tnl_cfg.dont_fragment) {
         smap_add(args, "df_default", "false");
     }
 
+    if (tnl_cfg.set_egress_pkt_mark) {
+        smap_add_format(args, "egress_pkt_mark",
+                        "%"PRIu32, tnl_cfg.egress_pkt_mark);
+    }
     return 0;
 }
 
@@ -770,11 +845,38 @@ get_stats(const struct netdev *netdev, struct netdev_stats *stats)
     return 0;
 }
 
+static enum netdev_pt_mode
+get_pt_mode(const struct netdev *netdev)
+{
+    struct netdev_vport *dev = netdev_vport_cast(netdev);
+
+    return dev->tnl_cfg.pt_mode;
+}
+
+
 
+#ifdef __linux__
+static int
+netdev_vport_get_ifindex(const struct netdev *netdev_)
+{
+    char buf[NETDEV_VPORT_NAME_BUFSIZE];
+    const char *name = netdev_vport_get_dpif_port(netdev_, buf, sizeof(buf));
+
+    return linux_get_ifindex(name);
+}
+
+#define NETDEV_VPORT_GET_IFINDEX netdev_vport_get_ifindex
+#define NETDEV_FLOW_OFFLOAD_API LINUX_FLOW_OFFLOAD_API
+#else /* !__linux__ */
+#define NETDEV_VPORT_GET_IFINDEX NULL
+#define NETDEV_FLOW_OFFLOAD_API NO_OFFLOAD_API
+#endif /* __linux__ */
+
 #define VPORT_FUNCTIONS(GET_CONFIG, SET_CONFIG,             \
                         GET_TUNNEL_CONFIG, GET_STATUS,      \
                         BUILD_HEADER,                       \
-                        PUSH_HEADER, POP_HEADER)            \
+                        PUSH_HEADER, POP_HEADER,            \
+                        GET_IFINDEX)                        \
     NULL,                                                   \
     netdev_vport_run,                                       \
     netdev_vport_wait,                                      \
@@ -799,7 +901,7 @@ get_stats(const struct netdev *netdev, struct netdev_stats *stats)
     netdev_vport_get_etheraddr,                             \
     NULL,                       /* get_mtu */               \
     NULL,                       /* set_mtu */               \
-    NULL,                       /* get_ifindex */           \
+    GET_IFINDEX,                                            \
     NULL,                       /* get_carrier */           \
     NULL,                       /* get_carrier_resets */    \
     NULL,                       /* get_miimon */            \
@@ -807,6 +909,7 @@ get_stats(const struct netdev *netdev, struct netdev_stats *stats)
                                                             \
     NULL,                       /* get_features */          \
     NULL,                       /* set_advertisements */    \
+    get_pt_mode,                                            \
                                                             \
     NULL,                       /* set_policing */          \
     NULL,                       /* get_qos_types */         \
@@ -838,17 +941,21 @@ get_stats(const struct netdev *netdev, struct netdev_stats *stats)
     NULL,                   /* rx_dealloc */                \
     NULL,                   /* rx_recv */                   \
     NULL,                   /* rx_wait */                   \
-    NULL,                   /* rx_drain */
+    NULL,                   /* rx_drain */                  \
+                                                            \
+    NETDEV_FLOW_OFFLOAD_API
 
 
-#define TUNNEL_CLASS(NAME, DPIF_PORT, BUILD_HEADER, PUSH_HEADER, POP_HEADER)   \
+#define TUNNEL_CLASS(NAME, DPIF_PORT, BUILD_HEADER, PUSH_HEADER, POP_HEADER,   \
+                     GET_IFINDEX)                                              \
     { DPIF_PORT,                                                               \
         { NAME, false,                                                         \
           VPORT_FUNCTIONS(get_tunnel_config,                                   \
                           set_tunnel_config,                                   \
                           get_netdev_tunnel_config,                            \
                           tunnel_get_status,                                   \
-                          BUILD_HEADER, PUSH_HEADER, POP_HEADER) }}
+                          BUILD_HEADER, PUSH_HEADER, POP_HEADER,               \
+                          GET_IFINDEX) }}
 
 void
 netdev_vport_tunnel_register(void)
@@ -858,15 +965,18 @@ netdev_vport_tunnel_register(void)
     static const struct vport_class vport_classes[] = {
         TUNNEL_CLASS("geneve", "genev_sys", netdev_geneve_build_header,
                                             netdev_tnl_push_udp_header,
-                                            netdev_geneve_pop_header),
+                                            netdev_geneve_pop_header,
+                                            NULL),
         TUNNEL_CLASS("gre", "gre_sys", netdev_gre_build_header,
                                        netdev_gre_push_header,
-                                       netdev_gre_pop_header),
+                                       netdev_gre_pop_header,
+                                       NULL),
         TUNNEL_CLASS("vxlan", "vxlan_sys", netdev_vxlan_build_header,
                                            netdev_tnl_push_udp_header,
-                                           netdev_vxlan_pop_header),
-        TUNNEL_CLASS("lisp", "lisp_sys", NULL, NULL, NULL),
-        TUNNEL_CLASS("stt", "stt_sys", NULL, NULL, NULL),
+                                           netdev_vxlan_pop_header,
+                                           NETDEV_VPORT_GET_IFINDEX),
+        TUNNEL_CLASS("lisp", "lisp_sys", NULL, NULL, NULL, NULL),
+        TUNNEL_CLASS("stt", "stt_sys", NULL, NULL, NULL, NULL),
     };
     static struct ovsthread_once once = OVSTHREAD_ONCE_INITIALIZER;
 
@@ -893,6 +1003,6 @@ netdev_vport_patch_register(void)
               VPORT_FUNCTIONS(get_patch_config,
                               set_patch_config,
                               NULL,
-                              NULL, NULL, NULL, NULL) }};
+                              NULL, NULL, NULL, NULL, NULL) }};
     netdev_register_provider(&patch_class.netdev_class);
 }
