@@ -298,9 +298,11 @@ sflow_agent_get_counters(void *ds_, SFLPoller *poller,
 {
     struct dpif_sflow *ds = ds_;
     SFLCounters_sample_element elem, lacp_elem, of_elem, name_elem;
+    SFLCounters_sample_element eth_elem;
     enum netdev_features current;
     struct dpif_sflow_port *dsp;
     SFLIf_counters *counters;
+    SFLEthernet_counters* eth_counters;
     struct netdev_stats stats;
     enum netdev_flags flags;
     struct lacp_slave_stats lacp_stats;
@@ -343,14 +345,14 @@ sflow_agent_get_counters(void *ds_, SFLPoller *poller,
     counters->ifInOctets = stats.rx_bytes;
     counters->ifInUcastPkts = stats.rx_packets;
     counters->ifInMulticastPkts = stats.multicast;
-    counters->ifInBroadcastPkts = -1;
+    counters->ifInBroadcastPkts = stats.rx_broadcast_packets;
     counters->ifInDiscards = stats.rx_dropped;
     counters->ifInErrors = stats.rx_errors;
     counters->ifInUnknownProtos = -1;
     counters->ifOutOctets = stats.tx_bytes;
     counters->ifOutUcastPkts = stats.tx_packets;
-    counters->ifOutMulticastPkts = -1;
-    counters->ifOutBroadcastPkts = -1;
+    counters->ifOutMulticastPkts = stats.tx_multicast_packets;
+    counters->ifOutBroadcastPkts = stats.tx_broadcast_packets;
     counters->ifOutDiscards = stats.tx_dropped;
     counters->ifOutErrors = stats.tx_errors;
     counters->ifPromiscuousMode = 0;
@@ -407,6 +409,25 @@ sflow_agent_get_counters(void *ds_, SFLPoller *poller,
       (OVS_FORCE uint32_t)dsp->ofport->ofp_port;
     SFLADD_ELEMENT(cs, &of_elem);
 
+    /* Include ethernet counters */
+    memset(&eth_elem, 0, sizeof eth_elem);
+    eth_elem.tag = SFLCOUNTERS_ETHERNET;
+    eth_counters = &eth_elem.counterBlock.ethernet;
+    eth_counters->dot3StatsAlignmentErrors = stats.rx_frame_errors;
+    eth_counters->dot3StatsFCSErrors = stats.rx_crc_errors;
+    eth_counters->dot3StatsFrameTooLongs = stats.rx_oversize_errors;
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsSingleCollisionFrames);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsMultipleCollisionFrames);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsSQETestErrors);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsDeferredTransmissions);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsLateCollisions);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsExcessiveCollisions);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsInternalMacTransmitErrors);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsCarrierSenseErrors);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsInternalMacReceiveErrors);
+    SFL_UNDEF_COUNTER(eth_counters->dot3StatsSymbolErrors);
+    SFLADD_ELEMENT(cs, &eth_elem);
+
     sfl_poller_writeCountersSample(poller, cs);
 }
 
@@ -449,7 +470,10 @@ sflow_choose_agent_address(const char *agent_device,
             struct in6_addr addr6, src, gw;
 
             in6_addr_set_mapped_ipv4(&addr6, sa.sin.sin_addr.s_addr);
-            if (ovs_router_lookup(&addr6, name, &src, &gw)) {
+            /* sFlow only supports target in default routing table with
+             * packet mark zero.
+             */
+            if (ovs_router_lookup(0, &addr6, name, &src, &gw)) {
 
                 in4.s_addr = in6_addr_get_mapped_ipv4(&src);
                 goto success;
@@ -877,7 +901,7 @@ sflow_read_tnl_push_action(const struct nlattr *attr,
     const struct ip_header *ip
         = ALIGNED_CAST(const struct ip_header *, eth + 1);
 
-    sflow_actions->out_port = u32_to_odp(data->out_port);
+    sflow_actions->out_port = data->out_port;
 
     /* Ethernet. */
     /* TODO: SFlow does not currently define a MAC-in-MAC
@@ -1022,7 +1046,11 @@ sflow_read_set_action(const struct nlattr *attr,
     case OVS_KEY_ATTR_CT_ZONE:
     case OVS_KEY_ATTR_CT_MARK:
     case OVS_KEY_ATTR_CT_LABELS:
+    case OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV4:
+    case OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV6:
     case OVS_KEY_ATTR_UNSPEC:
+    case OVS_KEY_ATTR_PACKET_TYPE:
+    case OVS_KEY_ATTR_NSH:
     case __OVS_KEY_ATTR_MAX:
     default:
         break;
@@ -1132,6 +1160,7 @@ dpif_sflow_read_actions(const struct flow *flow,
 	case OVS_ACTION_ATTR_RECIRC:
 	case OVS_ACTION_ATTR_HASH:
         case OVS_ACTION_ATTR_CT:
+        case OVS_ACTION_ATTR_METER:
 	    break;
 
 	case OVS_ACTION_ATTR_SET_MASKED:
@@ -1161,7 +1190,17 @@ dpif_sflow_read_actions(const struct flow *flow,
 	    dpif_sflow_pop_mpls_lse(sflow_actions);
 	    break;
 	}
+	case OVS_ACTION_ATTR_PUSH_ETH:
+	case OVS_ACTION_ATTR_POP_ETH:
+	    /* TODO: SFlow does not currently define a MAC-in-MAC
+	     * encapsulation structure.  We could use an extension
+	     * structure to report this.
+	     */
+	    break;
 	case OVS_ACTION_ATTR_SAMPLE:
+	case OVS_ACTION_ATTR_CLONE:
+        case OVS_ACTION_ATTR_ENCAP_NSH:
+        case OVS_ACTION_ATTR_DECAP_NSH:
 	case OVS_ACTION_ATTR_UNSPEC:
 	case __OVS_ACTION_ATTR_MAX:
 	default:
@@ -1266,8 +1305,8 @@ dpif_sflow_received(struct dpif_sflow *ds, const struct dp_packet *packet,
     /* Add extended switch element. */
     memset(&switchElem, 0, sizeof(switchElem));
     switchElem.tag = SFLFLOW_EX_SWITCH;
-    switchElem.flowType.sw.src_vlan = vlan_tci_to_vid(flow->vlan_tci);
-    switchElem.flowType.sw.src_priority = vlan_tci_to_pcp(flow->vlan_tci);
+    switchElem.flowType.sw.src_vlan = vlan_tci_to_vid(flow->vlans[0].tci);
+    switchElem.flowType.sw.src_priority = vlan_tci_to_pcp(flow->vlans[0].tci);
 
     /* Retrieve data from user_action_cookie. */
     vlan_tci = cookie->sflow.vlan_tci;
