@@ -68,7 +68,8 @@ static void consider_logical_flow(struct controller_ctx *ctx,
                                   uint32_t *conj_id_ofs,
                                   const struct shash *addr_sets,
                                   struct hmap *flow_table,
-                                  struct sset *active_tunnels);
+                                  struct sset *active_tunnels,
+                                  struct sset *local_lport_ids);
 
 static bool
 lookup_port_cb(const void *aux_, const char *port_name, unsigned int *portp)
@@ -145,7 +146,8 @@ add_logical_flows(struct controller_ctx *ctx,
                   const struct sbrec_chassis *chassis,
                   const struct shash *addr_sets,
                   struct hmap *flow_table,
-                  struct sset *active_tunnels)
+                  struct sset *active_tunnels,
+                  struct sset *local_lport_ids)
 {
     uint32_t conj_id_ofs = 1;
     const struct sbrec_logical_flow *lflow;
@@ -170,7 +172,8 @@ add_logical_flows(struct controller_ctx *ctx,
                               lflow, local_datapaths,
                               group_table, chassis,
                               &dhcp_opts, &dhcpv6_opts, &conj_id_ofs,
-                              addr_sets, flow_table, active_tunnels);
+                              addr_sets, flow_table, active_tunnels,
+                              local_lport_ids);
     }
 
     dhcp_opts_destroy(&dhcp_opts);
@@ -189,7 +192,8 @@ consider_logical_flow(struct controller_ctx *ctx,
                       uint32_t *conj_id_ofs,
                       const struct shash *addr_sets,
                       struct hmap *flow_table,
-                      struct sset *active_tunnels)
+                      struct sset *active_tunnels,
+                      struct sset *local_lport_ids)
 {
     /* Determine translation of logical table IDs to physical table IDs. */
     bool ingress = !strcmp(lflow->pipeline, "ingress");
@@ -301,6 +305,19 @@ consider_logical_flow(struct controller_ctx *ctx,
         if (m->match.wc.masks.conj_id) {
             m->match.flow.conj_id += *conj_id_ofs;
         }
+        if (is_switch(ldp)) {
+            unsigned int reg_index
+                = (ingress ? MFF_LOG_INPORT : MFF_LOG_OUTPORT) - MFF_REG0;
+            int64_t port_id = m->match.flow.regs[reg_index];
+            if (port_id) {
+                int64_t dp_id = lflow->logical_datapath->tunnel_key;
+                char buf[16];
+                snprintf(buf, sizeof(buf), "%"PRId64"_%"PRId64, dp_id, port_id);
+                if (!sset_contains(local_lport_ids, buf)) {
+                    continue;
+                }
+            }
+        }
         if (!m->n) {
             ofctrl_add_flow(flow_table, ptable, lflow->priority,
                             lflow->header_.uuid.parts[0], &m->match, &ofpacts);
@@ -413,11 +430,12 @@ lflow_run(struct controller_ctx *ctx,
           struct group_table *group_table,
           const struct shash *addr_sets,
           struct hmap *flow_table,
-          struct sset *active_tunnels)
+          struct sset *active_tunnels,
+          struct sset *local_lport_ids)
 {
     add_logical_flows(ctx, chassis_index, local_datapaths,
                       group_table, chassis, addr_sets, flow_table,
-                      active_tunnels);
+                      active_tunnels, local_lport_ids);
     add_neighbor_flows(ctx, flow_table);
 }
 
